from typing import List

import pytest
from watchdog import events

from pytest_watcher import watcher


@pytest.fixture
def trigger():
    return watcher.Trigger()


@pytest.mark.parametrize("event_type", watcher.EventHandler.EVENTS_WATCHED)
def test_event_types_watched(event_type, trigger: watcher.Trigger):
    event = events.FileSystemEvent("main.py")
    event.event_type = event_type

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert trigger.is_active()


def test_event_types_not_watched(trigger: watcher.Trigger):
    event = events.FileClosedEvent("main.py")

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert not trigger.is_active()


@pytest.mark.parametrize(
    "event_class", [events.FileSystemMovedEvent, events.FileMovedEvent]
)
def test_file_moved_dest_watched(event_class, trigger: watcher.Trigger):
    event = event_class("main.tmp", "main.py")

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert trigger.is_active()


@pytest.mark.parametrize(
    "event_class", [events.FileSystemMovedEvent, events.FileMovedEvent]
)
def test_file_moved_dest_not_watched(event_class, trigger: watcher.Trigger):
    event = event_class("main.tmp", "main.temp")

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert not trigger.is_active()


@pytest.mark.parametrize("path", ["main.py", "./main.py", "/home/project/main.py"])
def test_patterns_default_watched(trigger: watcher.Trigger, path: str):
    event = events.FileCreatedEvent(path)

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert trigger.is_active()


@pytest.mark.parametrize("path", ["main.pyc", "sqlite.db", "/home/project/file.txt"])
def test_patterns_default_not_watched(trigger: watcher.Trigger, path: str):
    event = events.FileCreatedEvent(path)

    handler = watcher.EventHandler(trigger)
    handler.dispatch(event)

    assert not trigger.is_active()


@pytest.mark.parametrize(
    ("patterns", "path"),
    [
        (["*.txt"], "file.txt"),
        (["main.pyc"], "main.pyc"),
        (["/home/path/example.txt"], "/home/path/example.txt"),
        (["*some*.txt"], "/home/path/something.txt"),
    ],
)
def test_patterns_custom_watched(
    trigger: watcher.Trigger, patterns: List[str], path: str
):
    event = events.FileCreatedEvent(path)

    handler = watcher.EventHandler(trigger, patterns=patterns)
    handler.dispatch(event)

    assert trigger.is_active()


@pytest.mark.parametrize(
    ("patterns", "path"),
    [
        (["*.txt"], "file.txtf"),
        (["*.pyi", "*.pdb"], "wrong.pdf"),
        (["/home/path/example.txt"], "/home/path/wrong.txt"),
    ],
)
def test_patterns_custom_not_watched(
    trigger: watcher.Trigger, patterns: List[str], path: str
):
    event = events.FileCreatedEvent(path)

    handler = watcher.EventHandler(trigger, patterns=patterns)
    handler.dispatch(event)

    assert not trigger.is_active()


@pytest.mark.parametrize(
    ("ignore_patterns", "path"),
    [
        (["ignore/*.py"], "ignore/myfile.py"),
        (["ignore/**"], "ignore/main.py"),
        (["*pytest*"], "/home/project/pytest.yaml"),
    ],
)
def test_patterns_ignore_not_watched(
    trigger: watcher.Trigger, ignore_patterns: List[str], path: str
):
    event = events.FileCreatedEvent(path)

    handler = watcher.EventHandler(trigger, ignore_patterns=ignore_patterns)
    handler.dispatch(event)

    assert not trigger.is_active()



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from pytest_watcher.config import Config
from pytest_watcher.terminal import Terminal
from pytest_watcher.trigger import Trigger


@pytest.fixture
def trigger():
    return Trigger()


@pytest.fixture()
def mock_subprocess_run(mocker: MockerFixture):
    return mocker.patch("pytest_watcher.watcher.subprocess.run", autospec=True)


@pytest.fixture(autouse=True)
def mock_time_sleep(mocker: MockerFixture):
    return mocker.patch("pytest_watcher.watcher.time.sleep", autospec=True)


@pytest.fixture
def mock_observer(mocker: MockerFixture):
    return mocker.patch("pytest_watcher.watcher.Observer", autospec=True)


@pytest.fixture
def mock_main_loop(mocker: MockerFixture):
    mock = mocker.patch("pytest_watcher.watcher.main_loop", autospec=True)
    mock.side_effect = InterruptedError
    return mock


@pytest.fixture(scope="session")
def tmp_path() -> Path:
    return Path("tests/tmp")


@pytest.fixture(scope="session", autouse=True)
def create_tmp_dir(tmp_path: Path):
    tmp_path.mkdir(exist_ok=True)


@pytest.fixture
def pyproject_toml_path(tmp_path: Path):
    path = tmp_path.joinpath("pyproject.toml")
    path.touch()

    yield path

    path.unlink()


@pytest.fixture
def config():
    return Config(path=Path())


@pytest.fixture
def mock_run_command(mocker: MockerFixture):
    return mocker.patch("pytest_watcher.commands.Manager.run_command", autospec=True)


@pytest.fixture
def mock_terminal(mocker: MockerFixture):
    return MagicMock(spec=Terminal)


@pytest.fixture(autouse=True)
def _patch_get_terminal(mocker: MockerFixture, mock_terminal: MagicMock):
    mocker.patch("pytest_watcher.watcher.get_terminal", mock_terminal)



from pytest_watcher import commands
from pytest_watcher.config import Config
from pytest_watcher.terminal import Terminal
from pytest_watcher.trigger import Trigger


def test_run_command(trigger: Trigger, config: Config, mock_terminal: Terminal):
    class DummyCommand(commands.Command):
        character = "0"
        caption = "0"
        description = "test"
        show_in_menu = False

        def __init__(self):
            self.invoked = False

        def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
            self.invoked = True

    command = commands.Manager.get_command("0")

    assert isinstance(command, DummyCommand)
    assert command.invoked is False

    commands.Manager.run_command("0", trigger, mock_terminal, config)

    assert command.invoked is True

    commands.Manager._registry.pop(DummyCommand.character)



from argparse import Namespace
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytest_watcher.config import (
    CLI_FIELDS,
    CONFIG_SECTION_NAME,
    find_config,
    parse_config,
)
from pytest_watcher.constants import DEFAULT_DELAY
from pytest_watcher.watcher import Config


@pytest.fixture(autouse=True)
def _patch_cwd(mocker: MockerFixture, tmp_path: Path):
    mock = mocker.patch("pytest_watcher.config.Path.cwd")
    mock.return_value = tmp_path


@pytest.fixture
def empty_namespace(tmp_path: Path):
    return Namespace(
        path=tmp_path,
        now=None,
        clear=None,
        delay=None,
        runner=None,
        patterns=None,
        ignore_patterns=None,
    )


@pytest.fixture
def config(empty_namespace: Namespace) -> Config:
    return Config.create(empty_namespace)


@pytest.fixture
def pyproject_toml(pyproject_toml_path: Path) -> Path:
    pyproject_toml_path.write_text(
        f"[tool.{CONFIG_SECTION_NAME}]\n"
        "now = true\n"
        "delay = 999\n"
        "runner = 'tox'\n"
        "runner_args = ['--lf', '--nf']\n"
        "patterns = ['*.py', '.env']\n"
        "ignore_patterns = ['ignore.py']\n"
    )

    return pyproject_toml_path


@pytest.fixture
def namespace(tmp_path: Path) -> Namespace:
    return Namespace(
        path=tmp_path,
        now=True,
        clear=True,
        delay=20,
        runner="tox",
        patterns=["*.py", ".env"],
        ignore_patterns=["main.py"],
    )


def test_default_values(config: Config):
    assert config.now is False
    assert config.delay == DEFAULT_DELAY
    assert config.runner == "pytest"
    assert config.runner_args == []
    assert config.patterns == []
    assert config.ignore_patterns == []


def test_cli_args(namespace: Namespace, tmp_path: Path):
    runner_args = ["--lf", "--nf"]

    config = Config.create(namespace=namespace, extra_args=runner_args)

    for f in CLI_FIELDS:
        assert getattr(config, f) == getattr(namespace, f)

    assert config.runner_args == runner_args


def test_cli_args_none_values_are_skipped(tmp_path: Path):
    namespace = Namespace(
        path=tmp_path,
        now=None,
        clear=None,
        delay=None,
        runner=None,
        patterns=None,
        ignore_patterns=None,
    )

    config = Config.create(namespace=namespace, extra_args=None)

    for f in CLI_FIELDS:
        assert getattr(config, f) is not None

    assert config.runner_args == []


def test_pyproject_toml(pyproject_toml: Path, config: Config):
    assert config.now is True
    assert config.delay == 999
    assert config.runner == "tox"
    assert config.runner_args == ["--lf", "--nf"]
    assert config.patterns == ["*.py", ".env"]
    assert config.ignore_patterns == ["ignore.py"]


def test_cli_args_preferred_over_pyproject_toml(
    pyproject_toml: Path, namespace: Namespace
):
    extra_args = ["--cli", "--args"]

    config = Config.create(namespace, extra_args=extra_args)

    for f in CLI_FIELDS:
        assert getattr(config, f) == getattr(namespace, f)

    assert config.runner_args == extra_args


@pytest.mark.parametrize(
    ("work_dir"),
    [
        Path(""),
        Path("test/"),
        Path("test/nested/dir/"),
    ],
)
def test_find_config(
    tmp_path: Path, pyproject_toml_path: Path, work_dir: Path, mocker: MockerFixture
):
    work_dir = tmp_path.joinpath(work_dir)
    work_dir.mkdir(parents=True, exist_ok=True)

    got = find_config(tmp_path)
    assert got == pyproject_toml_path, "Config file not found"


def test_parse_config_no_section(pyproject_toml_path: Path):
    pyproject_toml_path.write_text("[tool.another_section]\ndelay = 2\nrunner = 'tox'\n")

    got = parse_config(pyproject_toml_path)

    assert got == {}


def test_parse_config_parse_error(pyproject_toml_path: Path):
    pyproject_toml_path.write_text(
        f"[tool.{CONFIG_SECTION_NAME}]\ndelay = 2\nrunner = invalid\n"
    )

    with pytest.raises(SystemExit, match="Error parsing pyproject.toml"):
        parse_config(pyproject_toml_path)


def test_parse_config_unrecognized_option(pyproject_toml_path: Path):
    pyproject_toml_path.write_text(f"[tool.{CONFIG_SECTION_NAME}]\nfoo = 'bar'\n")

    with pytest.raises(SystemExit, match="Unrecognized option"):
        parse_config(pyproject_toml_path)



import abc
import logging
import os
import select
import sys
from typing import List, Optional

try:
    import termios
    import tty
except ImportError:
    pass


class Terminal(abc.ABC):
    def clear(self):
        pass

    def print(self, msg: str) -> None:
        pass

    def print_header(self, runner_args: List[str]):
        self.print(f"[pytest-watcher]\nCurrent runner args: [{' '.join(runner_args)}]")

    def print_short_menu(self, runner_args: List[str]) -> None:
        self.print_header(runner_args)
        self.print("\nPress w to show menu")

    def print_menu(self, runner_args: List[str]) -> None:
        from . import commands

        self.print_header(runner_args)
        self.print("\n\nControls:\n")

        for command in commands.Manager.list_commands():
            if command.show_in_menu:
                self.print(f"> {command.caption.ljust(5)} : {command.description}\n")

    def enter_capturing_mode(self) -> None:
        pass

    def capture_keystroke(self) -> Optional[str]:
        pass

    def reset(self) -> None:
        pass


class PosixTerminal(Terminal):
    def __init__(self) -> None:
        self._initial_state = termios.tcgetattr(sys.stdin.fileno())

    def print(self, msg: str) -> None:
        sys.stdout.write(msg)

    def clear(self) -> None:
        sys.stdout.write("\033c")
        sys.stdout.flush()

    def enter_capturing_mode(self) -> None:
        sys.stdin.flush()
        tty.setcbreak(sys.stdin.fileno())

    def capture_keystroke(self) -> Optional[str]:
        if select.select([sys.stdin], [], [], 0)[0]:
            return sys.stdin.read(1)
        return None

    def reset(self) -> None:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self._initial_state)


class DummyTerminal(Terminal):
    pass


def get_terminal() -> Terminal:
    if os.name == "posix":
        try:
            return PosixTerminal()
        except Exception:
            logging.exception(
                "Unable to initialize terminal state\nInteractive mode is disabled"
            )
    return DummyTerminal()



from __future__ import annotations

import abc
import sys
from typing import Dict, Optional, Type

from .config import Config
from .terminal import Terminal
from .trigger import Trigger


class Manager:
    _registry: Dict[str, Command] = {}

    @classmethod
    def list_commands(cls):
        return cls._registry.values()

    @classmethod
    def get_command(cls, character: str) -> Optional[Command]:
        return cls._registry.get(character)

    @classmethod
    def register(cls, command: Type[Command]):
        if command.character in cls._registry:
            raise ValueError(f"Duplicate character {repr(command.character)}")

        cls._registry[command.character] = command()

    @classmethod
    def run_command(
        cls, character: str, trigger: Trigger, term: Terminal, config: Config
    ) -> None:
        command = cls.get_command(character)
        if command:
            command.run(trigger, term, config)


class Command(abc.ABC):
    character: str
    caption: str
    description: str
    show_in_menu: bool = True

    def __init_subclass__(cls, **kwargs) -> None:
        for field in ("character", "caption", "description"):
            if not hasattr(cls, field):
                raise NotImplementedError(f"{cls.__name__}: {field} not specified")

        super().__init_subclass__(**kwargs)
        Manager.register(cls)

    @abc.abstractmethod
    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        """
        Modify runner_args in-place if needed and return a bool indicating whether
        tests should be triggered instantly
        """


class OpenMenuCommand(Command):
    character = "w"
    caption = "w"
    description = "show menu"
    show_in_menu = False

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        term.clear()
        term.print_menu(config.runner_args)


class InvokeCommand(Command):
    character = "\n"
    caption = "Enter"
    description = "Invoke test runner"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        trigger.emit_now()


class ResetRunnerArgsCommand(Command):
    character = "r"
    caption = "r"
    description = "reset all runner args"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        config.runner_args.clear()
        trigger.emit_now()


class ChangeRunnerArgsCommand(Command):
    character = "c"
    caption = "c"
    description = "change runner args"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        term.reset()
        raw = input("\nEnter new runner args: ")
        new_args = raw.strip().split()
        config.runner_args.clear()
        config.runner_args.extend(new_args)
        trigger.emit_now()


class OnlyFailedCommand(Command):
    character = "f"
    caption = "f"
    description = "run only failed tests (--lf)"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        if "--lf" not in config.runner_args:
            config.runner_args.append("--lf")
        trigger.emit_now()


class PDBCommand(Command):
    character = "p"
    caption = "p"
    description = "drop to pdb on fail (--pdb)"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        if "--pdb" not in config.runner_args:
            config.runner_args.append("--pdb")
        trigger.emit_now()


class VerboseCommand(Command):
    character = "v"
    caption = "v"
    description = "increase verbosity (-v)"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        if "-v" not in config.runner_args:
            config.runner_args.append("-v")
        trigger.emit_now()


class EraseScreenCommand(Command):
    character = "e"
    caption = "e"
    description = "Erase terminal screen"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        term.clear()
        term.print_short_menu(config.runner_args)


class QuitCommand(Command):
    character = "q"
    caption = "q"
    description = "quit pytest-watcher"

    def run(self, trigger: Trigger, term: Terminal, config: Config) -> None:
        sys.exit(0)



import argparse
from pathlib import Path
from typing import List, Sequence, Tuple

from .constants import DEFAULT_DELAY, VERSION


def parse_arguments(args: Sequence[str]) -> Tuple[argparse.Namespace, List[str]]:
    def _parse_patterns(arg: str):
        return arg.split(",")

    parser = argparse.ArgumentParser(
        prog="pytest_watcher",
        description="""
            Watch the <path> for file changes and trigger the test runner (pytest).\n
            Additional arguments are passed directly to the test runner.
        """,
        allow_abbrev=False,
    )
    parser.add_argument("path", type=Path, help="The path to watch for file changes.")
    parser.add_argument(
        "--now", action="store_true", help="Trigger the test run immediately"
    )
    parser.add_argument(
        "--clear",
        action="store_true",
        help="Clear the terminal screen before test run",
    )
    parser.add_argument(
        "--delay",
        type=float,
        required=False,
        help="The delay (in seconds) before triggering "
        f"the test run (default: {DEFAULT_DELAY})",
    )
    parser.add_argument(
        "--runner",
        type=str,
        required=False,
        help="Specify the executable for running the tests (default: pytest)",
    )
    parser.add_argument(
        "--patterns",
        type=_parse_patterns,
        required=False,
        help="File patterns to watch, specified as comma-separated "
        "Unix-style patterns (default: '*.py')",
    )
    parser.add_argument(
        "--ignore-patterns",
        type=_parse_patterns,
        required=False,
        help="File patterns to ignore, specified as comma-separated "
        "Unix-style patterns (default: '')",
    )
    parser.add_argument("--version", action="version", version=VERSION)

    return parser.parse_known_args(args)



from pathlib import Path
from typing import List

import pytest

from pytest_watcher.watcher import VERSION, parse_arguments


@pytest.mark.parametrize(
    ("args", "path_to_watch"),
    [
        (["."], Path(".")),
        (["/project/"], Path("/project")),
        (["../project/"], Path("../project")),
        (["/project/", "/tests/"], Path("/project")),
    ],
)
def test_path(args: List[str], path_to_watch: str):
    parsed, _ = parse_arguments(args)

    assert parsed.path == path_to_watch


def test_delay():
    parsed, _ = parse_arguments([".", "--delay", "999"])
    assert parsed.delay == 999


def test_now():
    parsed, _ = parse_arguments([".", "--now"])
    assert parsed.now is True


def test_runner():
    parsed, _ = parse_arguments([".", "--runner", "tox"])
    assert parsed.runner == "tox"


def test_clear():
    parsed, _ = parse_arguments([".", "--clear"])
    assert parsed.clear is True


@pytest.mark.parametrize(
    ("args", "patterns"),
    [
        ([".", "--patterns", "*.py,*.env"], ["*.py", "*.env"]),
        (
            [".", "--patterns", "*.py,*.env,project/pyproject.toml"],
            ["*.py", "*.env", "project/pyproject.toml"],
        ),
    ],
)
def test_patterns(args: List[str], patterns: List[str]):
    parsed, _ = parse_arguments(args)
    assert parsed.patterns == patterns


@pytest.mark.parametrize(
    ("args", "ignore_patterns"),
    [
        (
            [".", "--ignore-patterns", "long-long-long-path,templates/*.py"],
            ["long-long-long-path", "templates/*.py"],
        ),
    ],
)
def test_ignore_patterns(args: List[str], ignore_patterns: List[str]):
    parsed, _ = parse_arguments(args)
    assert parsed.ignore_patterns == ignore_patterns


@pytest.mark.parametrize(
    ("args", "runner_args"),
    [
        (["."], []),
        ([".", "--lf", "--nf", "-vv"], ["--lf", "--nf", "-vv"]),
        ([".", "--runner", "tox", "--lf", "--nf", "-vv"], ["--lf", "--nf", "-vv"]),
        ([".", "--lf", "--nf", "-vv", "--runner", "tox"], ["--lf", "--nf", "-vv"]),
        (
            [".", "--ignore", "tests/test_watcher.py"],
            ["--ignore", "tests/test_watcher.py"],
        ),
    ],
)
def test_runner_args(args: List[str], runner_args: List[str]):
    _, parsed_args = parse_arguments(args)
    assert parsed_args == runner_args


def test_version(capsys: pytest.CaptureFixture):
    with pytest.raises(SystemExit):
        parse_arguments(["--version"])

    captured = capsys.readouterr()
    assert captured.out == f"{VERSION}\n"



import threading
import time


class Trigger:
    _value: float
    _lock: threading.Lock

    def __init__(self, delay: float = 0.0):
        self._lock = threading.Lock()
        self._value = 0
        self._delay = delay

    def emit(self):
        with self._lock:
            self._value = time.time() + self._delay

    def emit_now(self):
        with self._lock:
            self._value = time.time()

    def is_active(self):
        return self._value != 0

    def release(self):
        with self._lock:
            self._value = 0

    def check(self):
        return self._value > 0 and time.time() > self._value



from __future__ import annotations

import logging
import subprocess
import sys
import time

from watchdog.observers import Observer

from . import commands
from .config import Config
from .constants import LOOP_DELAY, VERSION
from .event_handler import EventHandler
from .parse import parse_arguments
from .terminal import Terminal, get_terminal
from .trigger import Trigger

logging.basicConfig(level=logging.INFO, format="[ptw] %(message)s")


def main_loop(trigger: Trigger, config: Config, term: Terminal) -> None:
    if trigger.check():
        term.reset()

        if config.clear:
            term.clear()

        try:
            subprocess.run([config.runner, *config.runner_args])

        finally:
            term.enter_capturing_mode()

        term.print_short_menu(config.runner_args)

        trigger.release()

    key = term.capture_keystroke()
    if key:
        commands.Manager.run_command(key, trigger, term, config)

    time.sleep(LOOP_DELAY)


def run():
    term = get_terminal()
    trigger = Trigger()

    namespace, runner_args = parse_arguments(sys.argv[1:])

    config = Config.create(namespace=namespace, extra_args=runner_args)

    event_handler = EventHandler(
        trigger, patterns=config.patterns, ignore_patterns=config.ignore_patterns
    )

    observer = Observer()
    observer.schedule(event_handler, config.path, recursive=True)
    observer.start()

    _print_intro(config)

    term.enter_capturing_mode()

    if config.now:
        trigger.emit()
    else:
        term.print_menu(config.runner_args)

    try:
        while True:
            main_loop(trigger, config, term)
    finally:
        observer.stop()
        observer.join()

        term.reset()


def _print_intro(config: Config) -> None:
    sys.stdout.write(f"pytest-watcher version {VERSION}\n")
    sys.stdout.write(f"Runner command: {config.runner}\n")
    sys.stdout.write(f"Waiting for file changes in {config.path.absolute()}\n")



from argparse import Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Mapping, Optional

from .constants import DEFAULT_DELAY

try:
    import tomllib
except ImportError:
    import tomli as tomllib


CONFIG_SECTION_NAME = "pytest-watcher"
CLI_FIELDS = {"now", "clear", "delay", "runner", "patterns", "ignore_patterns"}
CONFIG_FIELDS = CLI_FIELDS | {"runner_args"}


@dataclass
class Config:
    path: Path
    now: bool = False
    clear: bool = False
    delay: float = DEFAULT_DELAY
    runner: str = "pytest"
    runner_args: List[str] = field(default_factory=list)
    patterns: List[str] = field(default_factory=list)
    ignore_patterns: List[str] = field(default_factory=list)

    @classmethod
    def create(
        cls, namespace: Namespace, extra_args: Optional[List[str]] = None
    ) -> "Config":
        instance = cls(path=namespace.path)

        config_path = find_config(namespace.path)
        if config_path:
            parsed = parse_config(config_path)
            instance._update_from_mapping(parsed)

        instance._update_from_namespace(namespace, extra_args or [])
        return instance

    def _update_from_mapping(self, data: Mapping):
        for key, val in data.items():
            setattr(self, key, val)

    def _update_from_namespace(
        self, namespace: Namespace, runner_args: Optional[List[str]]
    ):
        self.path = namespace.path

        for f in CLI_FIELDS:
            val = getattr(namespace, f)
            if val:
                setattr(self, f, val)

        if runner_args:
            self.runner_args = runner_args


def find_config(cwd: Path) -> Optional[Path]:
    filename = "pyproject.toml"

    for path in (cwd, *cwd.parents):
        config_path = path.joinpath(filename)

        if config_path.exists():
            return config_path

    return None


def parse_config(path: Path) -> Mapping:
    with open(path, "rb") as f:
        try:
            data = tomllib.load(f)
        except Exception as exc:
            raise SystemExit(f"Error parsing pyproject.toml\n{exc}")

    try:
        data = data["tool"][CONFIG_SECTION_NAME]
    except KeyError:
        return {}

    for key in data.keys():
        if key not in CONFIG_FIELDS:
            raise SystemExit(
                f"Error parsing pyproject.toml.\nUnrecognized option: {key}"
            )
    return data



import sys
from pathlib import Path
from unittest.mock import MagicMock, sentinel

import pytest
from freezegun import freeze_time
from pytest_mock.plugin import MockerFixture

from pytest_watcher import watcher
from pytest_watcher.config import Config
from pytest_watcher.constants import LOOP_DELAY
from pytest_watcher.terminal import Terminal
from pytest_watcher.trigger import Trigger


@freeze_time("2020-01-01 00:00:00")
def test_main_loop_does_not_invoke_runner_without_trigger(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    config: Config,
    mock_terminal: Terminal,
    trigger: Trigger,
):
    watcher.main_loop(trigger, config, mock_terminal)

    mock_subprocess_run.assert_not_called()
    mock_time_sleep.assert_called_once_with(LOOP_DELAY)


@freeze_time("2020-01-01 00:00:00")
def test_main_loop_does_not_invoke_runner_before_delay(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    config: Config,
    mock_terminal: MagicMock,
    trigger: Trigger,
):
    trigger = Trigger(delay=5)
    trigger.emit()

    with freeze_time("2020-01-01 00:00:04"):
        watcher.main_loop(trigger, config, mock_terminal)

    mock_subprocess_run.assert_not_called()
    mock_time_sleep.assert_called_once_with(LOOP_DELAY)

    assert trigger.is_active()


@freeze_time("2020-01-01 00:00:00")
def test_main_loop_invokes_runner_after_delay(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    config: Config,
    mock_terminal: MagicMock,
):
    trigger = Trigger(delay=5)
    trigger.emit()

    config.runner = "custom"
    config.runner_args = ["foo", "bar"]

    with freeze_time("2020-01-01 00:00:06"):
        watcher.main_loop(trigger, config, mock_terminal)

    mock_subprocess_run.assert_called_once_with(["custom", "foo", "bar"])
    mock_time_sleep.assert_called_once_with(LOOP_DELAY)

    assert not trigger.is_active()


@freeze_time("2020-01-01 00:00:00")
def test_main_loop_clear(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    config: Config,
    mock_terminal: MagicMock,
    trigger: Trigger,
):
    config.clear = True
    trigger.emit()

    with freeze_time("2020-01-01 00:00:06"):
        watcher.main_loop(trigger, config, mock_terminal)

    mock_terminal.clear.assert_called_once_with()


@freeze_time("2020-01-01 00:00:00")
def test_main_loop_no_clear(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    config: Config,
    mock_terminal: MagicMock,
    trigger: Trigger,
):
    config.clear = False
    trigger.emit()

    with freeze_time("2020-01-01 00:00:06"):
        watcher.main_loop(trigger, config, mock_terminal)

    mock_terminal.clear.assert_not_called()


def test_main_loop_keystroke(
    mock_subprocess_run: MagicMock,
    mock_time_sleep: MagicMock,
    mock_run_command: MagicMock,
    config: Config,
    trigger: Trigger,
    mock_terminal: MagicMock,
):
    trigger.emit()
    mock_terminal.capture_keystroke.return_value = sentinel.KEYSTROKE

    watcher.main_loop(trigger, config, mock_terminal)

    mock_run_command.assert_called_once_with(
        sentinel.KEYSTROKE, trigger, mock_terminal, config
    )


def assert_observer_started(mock_observer: MagicMock, expected_path: Path):
    mock_observer.assert_called_once_with()
    observer_instance = mock_observer.return_value
    observer_instance.schedule.assert_called_once()
    observer_instance.start.assert_called_once()

    path = mock_observer.return_value.schedule.call_args[0][1]
    assert path == expected_path


def test_run_starts_the_observer_and_main_loop(
    mocker: MockerFixture,
    mock_observer: MagicMock,
    mock_main_loop: MagicMock,
):
    args = ["ptw", ".", "--lf", "--nf"]
    mocker.patch.object(sys, "argv", args)

    with pytest.raises(InterruptedError):
        watcher.run()

    assert_observer_started(mock_observer, Path("."))
    mock_main_loop.assert_called_once()


def test_run_invokes_tests_right_away_if_now_flag_is_set(
    mocker: MockerFixture,
    mock_observer: MagicMock,
    mock_main_loop: MagicMock,
):
    mock_emit = mocker.patch("pytest_watcher.watcher.Trigger.emit", autospec=True)

    args = ["ptw", ".", "--lf", "--nf", "--now"]

    mocker.patch.object(sys, "argv", args)

    with pytest.raises(InterruptedError):
        watcher.run()

    mock_emit.assert_called_once()


def test_patterns_and_ignore_patterns_are_passed_to_event_handler(
    mocker: MockerFixture,
    mock_observer: MagicMock,
    mock_main_loop: MagicMock,
):
    args = ["ptw", ".", "--patterns", "*.py,.env", "--ignore-patterns", "settings.py"]

    mocker.patch.object(sys, "argv", args)

    with pytest.raises(InterruptedError):
        watcher.run()

    event_handler = mock_observer.return_value.schedule.call_args[0][0]

    assert event_handler.patterns == ["*.py", ".env"]
    assert event_handler.ignore_patterns == ["settings.py"]



from __future__ import annotations

import logging
from typing import List, Optional

from watchdog import events
from watchdog.utils.patterns import match_any_paths

from .trigger import Trigger

trigger = Trigger()


class EventHandler:
    EVENTS_WATCHED = {
        events.EVENT_TYPE_CREATED,
        events.EVENT_TYPE_DELETED,
        events.EVENT_TYPE_MODIFIED,
        events.EVENT_TYPE_MOVED,
    }

    def __init__(
        self,
        trigger: Trigger,
        patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ):
        self._patterns = patterns or ["*.py"]
        self._ignore_patterns = ignore_patterns or []
        self._trigger = trigger

    @property
    def patterns(self) -> List[str]:
        return self._patterns

    @property
    def ignore_patterns(self) -> List[str]:
        return self._ignore_patterns

    def _is_event_watched(self, event: events.FileSystemEvent) -> bool:
        if event.event_type not in self.EVENTS_WATCHED:
            return False

        paths = [event.src_path]
        if hasattr(event, "dest_path"):
            # For file moved type events we are also interested in the destination
            paths.append(event.dest_path)

        return match_any_paths(
            paths,
            included_patterns=self.patterns,
            excluded_patterns=self.ignore_patterns,
        )

    def dispatch(self, event: events.FileSystemEvent) -> None:
        if self._is_event_watched(event):
            self._trigger.emit()
            logging.info(f"{event.src_path} {event.event_type}")
        else:
            logging.debug(f"IGNORED event: {event.event_type} src: {event.src_path}")
